# Generated by Django 3.2.3 on 2021-05-19 08:31

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("durin", "0002_client_throttlerate"),
    ]

    operations = [
        migrations.CreateModel(
            name="ClientSettings",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("description", models.TextField()),
                (
                    "client",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="settings",
                        to="durin.client",
                    ),
                ),
            ],
        ),
    ]



import binascii
from os import urandom

import humanize
from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from durin.settings import durin_settings
from durin.signals import token_renewed
from durin.throttling import UserClientRateThrottle

User = settings.AUTH_USER_MODEL


def _create_token_string() -> str:
    return binascii.hexlify(
        urandom(int(durin_settings.TOKEN_CHARACTER_LENGTH / 2))
    ).decode()


class Client(models.Model):
    """
    Identifier to represent any API client/browser that consumes your RESTful API.

    See ``example_project.models.ClientSettings``
    if you wish to extend this model per your convenience.
    """

    #: A unique identification name for the client.
    name = models.CharField(
        max_length=64,
        null=False,
        blank=False,
        db_index=True,
        unique=True,
        help_text=_("A unique identification name for the client."),
    )

    #: Token Time To Live (TTL) in timedelta. Format: ``DAYS HH:MM:SS``.
    token_ttl = models.DurationField(
        null=False,
        default=durin_settings.DEFAULT_TOKEN_TTL,
        verbose_name=_("Token Time To Live (TTL)"),
        help_text=_(
            """
            Token Time To Live (TTL) in timedelta. Format: <code>DAYS HH:MM:SS</code>.
            """
        ),
    )

    #: Throttle rate for requests authed with this client.
    #:
    #: **Format**: ``number_of_requests/period``
    #: where period should be one of: *('s', 'm', 'h', 'd')*.
    #: (same format as DRF's throttle rates)
    #:
    #: **Example**: ``100/h`` implies 100 requests each hour.
    #:
    #: .. versionadded:: 0.2
    throttle_rate = models.CharField(
        max_length=64,
        default="",
        blank=True,
        verbose_name=_("Throttle rate for requests authed with this client"),
        help_text=_(
            """Follows the same format as DRF's throttle rates.
            Format: <em>'number_of_requests/period'</em>
            where period should be one of: ('s', 'm', 'h', 'd').
            Example: '100/h' implies 100 requests each hour.
            """
        ),
        validators=[UserClientRateThrottle.validate_client_throttle_rate],
    )

    def __str__(self):
        td = humanize.naturaldelta(self.token_ttl)
        rate = self.throttle_rate or "null"
        return "({0}: {1}, {2})".format(self.name, td, rate)


class AuthTokenManager(models.Manager):
    def create(self, user, client, delta_ttl=None):
        token = _create_token_string()

        if delta_ttl is not None:
            expiry = timezone.now() + delta_ttl
        else:
            expiry = timezone.now() + client.token_ttl

        instance = super(AuthTokenManager, self).create(
            token=token, user=user, client=client, expiry=expiry
        )
        return instance


class AuthToken(models.Model):
    """
    Token model with a unique constraint on ``User`` <-> ``Client`` relationship.
    """

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["user", "client"], name="unique token for user per client"
            )
        ]

    objects = AuthTokenManager()

    #: Token string
    token = models.CharField(
        max_length=durin_settings.TOKEN_CHARACTER_LENGTH,
        null=False,
        blank=False,
        db_index=True,
        unique=True,
        help_text=_("Token is auto-generated on save."),
    )
    #: :class:`~User` ForeignKey
    user = models.ForeignKey(
        User,
        null=False,
        blank=False,
        related_name="auth_token_set",
        on_delete=models.CASCADE,
    )
    #: :class:`~Client` ForeignKey
    client = models.ForeignKey(
        Client,
        null=False,
        blank=False,
        related_name="auth_token_set",
        on_delete=models.CASCADE,
    )
    #: Created time
    created = models.DateTimeField(auto_now_add=True)
    #: Expiry time
    expiry = models.DateTimeField(null=False)

    def renew_token(self, request=None) -> "timezone.datetime":
        """
        Utility function to renew the token.

        Updates the :py:attr:`~expiry` attribute by ``Client.token_ttl``.
        """
        new_expiry = timezone.now() + self.client.token_ttl
        self.expiry = new_expiry
        self.save(update_fields=("expiry",))
        token_renewed.send(
            sender=self,
            request=request,
            new_expiry=new_expiry,
        )
        return new_expiry

    @property
    def expires_in(self) -> str:
        """
        Dynamic property that gives the :py:attr:`~expiry`
        attribute in human readable string format.

        Uses `humanize package <https://github.com/jmoiron/humanize>`__.
        """
        if self.expiry:
            td = self.expiry - self.created
            return humanize.naturaldelta(td)
        else:
            return "N/A"

    @property
    def has_expired(self) -> bool:
        """
        Dynamic property that returns ``True`` if token has expired,
        otherwise ``False``.
        """
        return timezone.now() > self.expiry

    def __repr__(self) -> str:
        return "({0}, {1}/{2})".format(
            self.token, self.user.get_username(), self.client.name
        )

    def __str__(self) -> str:
        return self.token



from django.urls import include, path
from rest_framework import routers

from durin import views

router = routers.SimpleRouter(trailing_slash=False)
router.register(
    r"sessions/", views.TokenSessionsViewSet, basename="durin_tokensessions"
)


urlpatterns = [
    path("login/", views.LoginView.as_view(), name="durin_login"),
    path("refresh/", views.RefreshView.as_view(), name="durin_refresh"),
    path("logout/", views.LogoutView.as_view(), name="durin_logout"),
    path("logoutall/", views.LogoutAllView.as_view(), name="durin_logoutall"),
    path("apiaccess/", views.APIAccessTokenView.as_view(), name="durin_apiaccess"),
    # router URLs
    path("", include(router.urls)),
]



"""
Durin provides two *abstract* permission classes which
make use of the :class:`durin.models.Client`
model it offers.

You will need to subclass them and modify the
``allowed_clients_name`` or ``disallowed_clients_name`` property per your wish.

Then you may use them the same way as other
`DRF permissions <https://www.django-rest-framework.org/api-guide/permissions/>`__ or
activate them on all your views by adding
them to ``REST_FRAMEWORK["DEFAULT_PERMISSION_CLASSES"]``
under your app's ``settings.py``
"""

from rest_framework.permissions import BasePermission


class AllowSpecificClients(BasePermission):
    """
    Allows access to only specific clients.\n
    Should be used along with :doc:`auth`.
    """

    #: Include names of allowed clients to ``allowed_clients_name``.
    allowed_clients_name = ()

    def has_permission(self, request, view):
        if not request.auth:
            return False
        return request.auth.client.name in self.allowed_clients_name


class DisallowSpecificClients(BasePermission):
    """
    Restrict specific clients from making requests.\n
    Should be used along with :doc:`auth`.
    """

    #: Include names of disallowed clients to ``disallowed_clients_name``.
    disallowed_clients_name = ()

    def has_permission(self, request, view):
        if not request.auth:
            return False
        return request.auth.client.name not in self.disallowed_clients_name



from django.contrib import admin

from durin import models


@admin.register(models.AuthToken)
class AuthTokenAdmin(admin.ModelAdmin):
    """Django's ModelAdmin for AuthToken.\n
    In most cases, you would want to override this to make
    ``AuthTokenAdmin.raw_id_fields = ("user",)``
    """

    list_select_related = True

    list_display = (
        "token",
        "client",
        "user",
        "created",
        "expires_in",
    )
    list_filter = ("client__name", "user")

    readonly_fields = ("token", "expiry", "created", "expires_in")

    def get_fieldsets(self, request, obj=None):
        """
        Hook for specifying fieldsets.

        :meta private:
        """
        if not obj:
            return [
                (
                    "API Auth Token",
                    {
                        "fields": (
                            "user",
                            "client",
                        ),
                        "description": """
                    <h3>Token will be auto-generated on save.</h3>
                    <h3>Token will carry the same expiry as the
                     selected client's token TTL.</h3>
                """,
                    },
                ),
            ]
        else:
            return super().get_fieldsets(request, obj)

    def has_change_permission(self, request, obj=None):
        """
        :meta private:
        """
        return False

    def save_model(self, request, obj, form, change):
        """
        Given a model instance save it to the database.

        :meta private:
        """
        if not change:
            created_obj = models.AuthToken.objects.create(
                user=obj.user, client=obj.client
            )
            obj.pk = created_obj.pk
            obj.token = created_obj.token
            obj.expiry = created_obj.expiry
        else:
            super(AuthTokenAdmin, self).save_model(request, obj, form, change)


@admin.register(models.Client)
class ClientAdmin(admin.ModelAdmin):
    """
    Django's ModelAdmin for :class:`Client` model.
    """

    list_display = (
        "id",
        "name",
        "token_ttl",
        "throttle_rate",
    )



# Generated by Django 3.2.3 on 2021-05-19 08:25

import datetime

from django.db import migrations, models

import durin.throttling


class Migration(migrations.Migration):

    dependencies = [
        ("durin", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="client",
            name="throttle_rate",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Follows the same format as DRF's throttle rates.\n            Format: <em>'number_of_requests/period'</em>\n            where period should be one of: ('s', 'm', 'h', 'd').\n            Example: '100/h' implies 100 requests each hour.\n            ",
                max_length=64,
                validators=[
                    durin.throttling.UserClientRateThrottle.validate_client_throttle_rate
                ],
                verbose_name="Throttle rate for requests authed with this client",
            ),
        ),
        migrations.AlterField(
            model_name="authtoken",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="client",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="client",
            name="token_ttl",
            field=models.DurationField(
                default=datetime.timedelta(days=1),
                help_text="\n            Token Time To Live (TTL) in timedelta. Format: <code>DAYS HH:MM:SS</code>.\n            ",
                verbose_name="Token Time To Live (TTL)",
            ),
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-24 14:30

import datetime

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Client",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        db_index=True,
                        help_text="A unique identification name for the client.",
                        max_length=64,
                        unique=True,
                    ),
                ),
                (
                    "token_ttl",
                    models.DurationField(
                        default=datetime.timedelta(days=1),
                        help_text="\n            Token Time To Live (TTL) in timedelta. Format: <em>DAYS HH:MM:SS</em>.\n            ",
                        verbose_name="Token Time To Live (TTL)",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AuthToken",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "token",
                    models.CharField(
                        db_index=True,
                        help_text="Token is auto-generated on save.",
                        max_length=64,
                        unique=True,
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("expiry", models.DateTimeField()),
                (
                    "client",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="auth_token_set",
                        to="durin.client",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="auth_token_set",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="authtoken",
            constraint=models.UniqueConstraint(
                fields=("user", "client"), name="unique token for user per client"
            ),
        ),
    ]



"""
Durin provides a throttling class which make use of the
:class:`durin.models.Client` model it offers.

Usage is the same way as other
`DRF throttling classes <https://django-rest-framework.org/api-guide/throttling/>`__.

Example ``settings.py``::

        #...snip...
        REST_FRAMEWORK = {
            "DEFAULT_THROTTLE_CLASSES": ["durin.throttling.UserClientRateThrottle"],
            "DEFAULT_THROTTLE_RATES": {"user_per_client": "10/min"},
        }
        #...snip...

.. data:: "user_per_client"

    default ``scope`` for the :class:`UserClientRateThrottle` class.

    The rate defined here serves as the default rate incase the
    ``throttle_rate`` field on :class:`durin.models.Client` is ``null``.
"""

from django.core.exceptions import ValidationError as DjValidationError
from rest_framework.throttling import UserRateThrottle


class UserClientRateThrottle(UserRateThrottle):  # lgtm [py/missing-call-to-init]
    """
    Throttles requests by identifying the *authed* **user-client pair**.

    This is useful if you want to define different user throttle rates
    per :class:`durin.models.Client` instance.

    .. versionadded:: 0.2
    """

    #: Same as the default
    cache_format = "throttle_%(scope)s_%(ident)s"

    #: Scope for this throttle
    scope = "user_per_client"

    def __init__(self):
        # lgtm [py/missing-call-to-init]
        pass

    def allow_request(self, request, view):
        """
        The ``rate`` is set here because we need access to
        ``request`` object which is not available inside :py:meth:`~get_rate`.
        """
        if request.user.is_authenticated and hasattr(request, "_auth"):
            rate = request._auth.client.throttle_rate
            self.rate = rate if rate else self.get_rate()
        else:
            self.rate = self.get_rate()

        self.num_requests, self.duration = self.parse_rate(self.rate)

        return super().allow_request(request, view)

    def get_cache_key(self, request, view) -> str:
        if request.user.is_authenticated:
            # overwrite
            if hasattr(request, "_auth"):
                ident = self._get_user_client_ident(request)
            else:
                ident = request.user.pk
        else:
            ident = self.get_ident(request)

        return self.cache_format % {"scope": self.scope, "ident": ident}

    def _get_user_client_ident(self, request) -> str:
        """
        Identify the user-client pair making the request.
        (assumes that ``request._auth`` is set, see :py:meth:`~get_cache_key`).
        """
        user_pk = request._auth.user_id
        client_pk = request._auth.client_id
        ident = "user-{0}.client-{1}".format(user_pk, client_pk)
        return ident

    @staticmethod
    def validate_client_throttle_rate(rate):
        """
        Used for validating the :attr:`throttle_rate` field
        on :class:`durin.models.Client`.

        *For internal use only.*
        """
        TIME_PERIODS_MAP = {"s": 1, "m": 60, "h": 3600, "d": 86400}

        try:
            num, period = rate.split("/")
            return int(num), TIME_PERIODS_MAP[period]
        except KeyError:
            raise DjValidationError("invalid period '{0}'.".format(period))
        except Exception as e:
            raise DjValidationError(e)



from datetime import datetime

from django.contrib.auth.signals import user_logged_in, user_logged_out
from rest_framework import mixins, status
from rest_framework.authtoken.serializers import AuthTokenSerializer
from rest_framework.exceptions import NotFound, ValidationError
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.serializers import DateTimeField
from rest_framework.views import APIView
from rest_framework.viewsets import GenericViewSet

from .models import AuthToken, Client
from .serializers import APIAccessTokenSerializer, TokenSessionsSerializer
from .settings import durin_settings


class LoginView(APIView):
    """Durin's Login View.\n
    This view will return a JSON response when valid ``username``, ``password`` and
    (if not overwritten) ``client`` fields are POSTed to the view using
    form data or JSON.

    It uses the default serializer provided by
    Django-Rest-Framework (``rest_framework.authtoken.serializers.AuthTokenSerializer``)
    to validate the user credentials.

    It is possible to customize LoginView behaviour by overriding the following
    helper methods:
    """

    authentication_classes = []
    permission_classes = []

    def get_context(self):
        """
        to change the context passed to the ``UserSerializer``.
        """
        return {"request": self.request, "format": self.format_kwarg, "view": self}

    @staticmethod
    def validate_and_return_user(request):
        serializer = AuthTokenSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        return serializer.validated_data["user"]

    def get_client_obj(self, request) -> "Client":
        """
        To get and return the associated :class:`durin.models.Client` object.

        :raises rest_framework.exceptions.ValidationError
        """
        client_name = request.data.get("client", None)
        if not client_name:
            raise ValidationError({"detail": "No client specified."})

        try:
            return Client.objects.get(name=client_name)
        except Client.DoesNotExist:
            raise ValidationError({"detail": "No client with that name."})

    def get_token_obj(self, request, client: "Client") -> "AuthToken":
        """
        Flow used to return the :class:`durin.models.AuthToken` object.
        """
        try:
            # if a token for this user-client pair already exists,
            # we can just return it
            token = AuthToken.objects.get(user=request.user, client=client)
            if durin_settings.REFRESH_TOKEN_ON_LOGIN:
                self.renew_token(request=request, token=token)
        except AuthToken.DoesNotExist:
            # create new token
            token = AuthToken.objects.create(request.user, client)

        return token

    def renew_token(self, request, token: "AuthToken") -> None:
        """
        How to renew the token instance in case
        ``settings.REFRESH_TOKEN_ON_LOGIN`` is set to ``True``.
        """
        token.renew_token(request=request)

    @staticmethod
    def format_expiry_datetime(expiry: "datetime") -> str:
        """
        To format the expiry ``datetime`` object at your convenience.
        """
        datetime_format = durin_settings.EXPIRY_DATETIME_FORMAT
        return DateTimeField(format=datetime_format).to_representation(expiry)

    def get_user_serializer_class(self):
        """
        To change the class used for serializing the user.
        """
        return durin_settings.USER_SERIALIZER

    def get_post_response_data(self, request, token_obj: "AuthToken") -> dict:
        """
        Override this to return a fully customized payload.
        """
        UserSerializer = self.get_user_serializer_class()
        data = {
            "expiry": self.format_expiry_datetime(token_obj.expiry),
            "token": token_obj.token,
        }
        if UserSerializer is not None:
            data["user"] = UserSerializer(request.user, context=self.get_context()).data
        return data

    def post(self, request, *args, **kwargs):
        request.user = self.validate_and_return_user(request)
        client = self.get_client_obj(request)
        token_obj = self.get_token_obj(request, client)
        user_logged_in.send(
            sender=request.user.__class__, request=request, user=request.user
        )
        data = self.get_post_response_data(request, token_obj)
        return Response(data)


class RefreshView(APIView):
    """Durin's Refresh View\n
    This view accepts only a post request with an empty body.
    It responds to Durin Token Authentication. On a successful request,

    1. The given token's expiry is extended by it's associated
       :py:attr:`durin.models.Client.token_ttl`
       duration and a JSON object will be returned containing a single ``expiry``
       key as the new timestamp for when the token expires.

    2. :meth:`durin.signals.token_renewed` is called.
    """

    @staticmethod
    def format_expiry_datetime(expiry: "datetime") -> str:
        """
        To format the expiry ``datetime`` object at your convenience.
        """
        datetime_format = durin_settings.EXPIRY_DATETIME_FORMAT
        return DateTimeField(format=datetime_format).to_representation(expiry)

    def renew_token(self, request, token: "AuthToken") -> "datetime":
        """
        How to renew the token instance.
        """
        new_expiry = token.renew_token(request=request)
        return new_expiry

    def post(self, request, *args, **kwargs):
        auth_token = request._auth
        new_expiry = self.renew_token(request=request, token=auth_token)
        new_expiry_repr = self.format_expiry_datetime(new_expiry)
        return Response({"expiry": new_expiry_repr}, status=status.HTTP_200_OK)


class LogoutView(APIView):
    """Durin's Logout View.\n
    This view accepts only a post request with an empty body.
    It responds to Durin Token Authentication. On a successful request,

    1. The token used to authenticate is deleted from
       the database and can no longer be used to authenticate.

    2. :meth:`django.contrib.auth.signals.user_logged_out` is called.

    :returns: 204 (No content)
    """

    def post(self, request, *args, **kwargs):
        request._auth.delete()
        user_logged_out.send(
            sender=request.user.__class__, request=request, user=request.user
        )
        return Response(None, status=status.HTTP_204_NO_CONTENT)


class LogoutAllView(APIView):
    """Durin's LogoutAllView.\n
    This view accepts only a post request with an empty body. It responds to Durin Token
    Authentication.
    On a successful request,

    1. The token used to authenticate, and **all other tokens**
       registered to the same ``User`` account, are deleted from the
       system and can no longer be used to authenticate.

    2. :meth:`django.contrib.auth.signals.user_logged_out` is called.

    :returns: 204 (No content)
    """

    def post(self, request, *args, **kwargs):
        request.user.auth_token_set.all().delete()
        user_logged_out.send(
            sender=request.user.__class__, request=request, user=request.user
        )
        return Response(None, status=status.HTTP_204_NO_CONTENT)


class TokenSessionsViewSet(
    mixins.ListModelMixin, mixins.DestroyModelMixin, GenericViewSet
):
    """Durin's TokenSessionsViewSet.\n
    - Returns list of active sessions of authed user.
    - Only ``list()`` and ``delete()`` operations.

    .. versionadded:: 1.0.0
    """

    queryset = AuthToken.objects.select_related("client").all()
    serializer_class = TokenSessionsSerializer
    pagination_class = None

    def get_queryset(self):
        qs = super().get_queryset()
        # filter against authed user
        qs = qs.filter(user=self.request.user)
        # exclude session for the APIAccess session
        # if `API_ACCESS_EXCLUDE_FROM_SESSIONS` setting is True
        if durin_settings.API_ACCESS_EXCLUDE_FROM_SESSIONS:
            qs = qs.exclude(client__name=durin_settings.API_ACCESS_CLIENT_NAME)
        return qs

    def perform_destroy(self, instance):
        """
        Overwrite to prevent deletion of object
        against which current request was authed.
        """
        if instance.pk == self.request.auth.pk:
            raise ValidationError(
                "Can't delete token if the request is authed with it."
                "Use {} instead.".format(reverse("durin_logout"))
            )
        instance.delete()


class APIAccessTokenView(APIView):
    """Durin's APIAccessTokenView.\n
    - ``GET`` -> get token-client pair info
    - ``POST`` -> create and get token-client pair info
    - ``DELETE`` -> delete existing API access token

    .. versionadded:: 1.0.0
    """

    @property
    def client_name(self) -> str:
        client_name = getattr(durin_settings, "API_ACCESS_CLIENT_NAME", None)
        # verify/ asssert
        assert client_name, "setting `API_ACCESS_CLIENT_NAME` must be set to use this."
        return client_name

    def get_serializer(self, *args, **kwargs):
        return APIAccessTokenSerializer(
            *args,
            **kwargs,
            context={
                "request": self.request,
                "format": self.format_kwarg,
                "view": self,
                "client_name": self.client_name,
            },
        )

    def get_object(self):
        try:
            instance = AuthToken.objects.get(
                user__pk=self.request.user.pk,
                client__name=self.client_name,
            )
        except AuthToken.DoesNotExist:
            raise NotFound()

        return instance

    def get(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)

    def post(self, request):
        serializer = self.get_serializer(data={})
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    def delete(self, request):
        instance = self.get_object()
        instance.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)



from django.contrib.auth import get_user_model
from rest_framework import serializers as rfs

from .models import AuthToken, Client
from .settings import durin_settings

User = get_user_model()
username_field = User.USERNAME_FIELD if hasattr(User, "USERNAME_FIELD") else "username"


class UserSerializer(rfs.ModelSerializer):
    class Meta:
        model = User
        fields = (username_field,)


class TokenSessionsSerializer(rfs.ModelSerializer):
    """
    Used in :class:`durin.views.TokenSessionsViewSet`.

    .. versionadded:: 1.0.0
    """

    class Meta:
        model = AuthToken
        fields = [
            "id",
            "client",
            "created",
            "expiry",
            "has_expired",
            "is_current",
            "expires_in_str",
        ]
        read_only_fields = [
            "token",
            "client",
            "user",
            "created",
            "expiry",
            "expires_in_str",
        ]

    client = rfs.SlugRelatedField(
        slug_field="name",
        read_only=True,
    )
    expires_in_str = rfs.CharField(source="expires_in")
    is_current = rfs.SerializerMethodField()

    def get_is_current(self, obj: AuthToken) -> bool:
        request = self.context["request"]
        return obj.pk == request.auth.pk


class APIAccessTokenSerializer(rfs.ModelSerializer):
    """
    Used in :class:`durin.views.APIAccessTokenView`.

    .. versionadded:: 1.0.0
    """

    class Meta:
        model = AuthToken
        fields = [
            "client",
            "created",
            "expiry",
            "has_expired",
            "expires_in_str",
        ]
        read_only_fields = [
            "token",
            "client",
            "user",
            "created",
            "expiry",
            "expires_in_str",
        ]

    client = rfs.SlugRelatedField(
        slug_field="name",
        read_only=True,
    )
    expires_in_str = rfs.CharField(source="expires_in", read_only=True)

    def get_field_names(self, declared_fields, info):
        """
        :meta private:
        """
        fields = super().get_field_names(declared_fields, info)
        if (
            durin_settings.API_ACCESS_RESPONSE_INCLUDE_TOKEN
            or self.context["request"].method.upper() == "POST"
        ):
            # only send token on POST/create request, or
            # if `API_ACCESS_RESPONSE_INCLUDE_TOKEN` setting is True
            fields.append("token")
        return fields

    def create(self, validated_data):
        """
        :meta private:
        """
        user = self.context["request"].user
        client_name = self.context["client_name"]
        if AuthToken.objects.filter(user=user, client__name=client_name).exists():
            raise rfs.ValidationError("An API token was already issued to you.")

        validated_data["user"] = user
        validated_data["client"] = Client.objects.get(name=client_name)
        return super().create(validated_data)
